import numpy as np
from .mc_constraint import MCConstraint


class ConstrainSwapByBasis(MCConstraint):
    """
    Constraint that restricts swaps of atoms within a given basis.
    This constraint is intended to be used together with canonical
    Monte Carlo calculations where the trial moves consist of swapping
    two atoms.

    Parameters:

    atoms: Atoms object
        ASE Atoms object used in the MC simulation

    index_by_basis: list
        Indices ordered by basis (same as `index_by_basis` parameter in
        `ClusterExpansionSettings`).
        If an Atoms object has 10 sites where the first 4 belongs to the
        first basis, the next 3 belongs to the next basis and the last 3
        belongs to the last basis, the `index_by_basis` would be
        [[0, 1, 2, 3], [4, 5, 6], [7, 8, 9]]

    Note: swaps are only allowed within each basis, not across two basis.
    """

    def __init__(self, atoms, index_by_basis):
        self.basis = np.zeros(len(atoms))

        for i, indices in enumerate(index_by_basis):
            for x in indices:
                self.basis[x] = i

    def __call__(self, system_changes):
        if len(system_changes) != 2:
            raise ValueError("ConstrainSwapByBasis requires swap move")

        i1 = system_changes[0].index
        b1 = self.basis[i1]
        i2 = system_changes[1].index
        b2 = self.basis[i2]
        return b1 == b2



import click
from clease.settings import settings_from_json
from . import main


@main.clease_cli.command()
@click.argument(
    "filename",
    type=click.Path(exists=True),
)
@click.option("--gui", "-g", is_flag=True, help="Open clusters in an ASE GUI?")
@click.option(
    "--no-table",
    "-t",
    is_flag=True,
    default=False,
    help="Disable printing the cluster table.",
)
def clusters(filename, gui, no_table):
    """Print the clusters that a settings object (saved as a JSON file) produces.

    FILENAME is the name of the settings JSON file
    """
    settings = settings_from_json(filename)
    if no_table is False:
        table = settings.clusters_table()
        click.echo(table)

    if gui:
        # We use ase.visualize.view here, to have the GUI spawned in
        # a background process, rather than using the settings.view_clusters()
        # method.
        # pylint: disable=import-outside-toplevel
        from ase.visualize import view

        figures = settings.get_all_figures_as_atoms()
        view(figures)



# pylint: disable=undefined-variable

from .system_changes import *
from .four_vector import *
from .trans_matrix import *
from .figure import *
from .mc_step import *

ADDITIONAL = ("system_changes",)
__all__ = (
    system_changes.__all__
    + four_vector.__all__
    + figure.__all__
    + trans_matrix.__all__
    + mc_step.__all__
    + ADDITIONAL
)



import time
import logging
import json
from copy import deepcopy
import numpy as np
from ase.units import kB
from .constraints import CollectiveVariableConstraint

logger = logging.getLogger(__name__)


class NoneNotAcceptedError(Exception):
    pass


class PeakNotAcceptedError(Exception):
    pass


class MetaDynamicsSampler:
    """
    Class for performing meta dynamics sampler

    Parameters:

    mc: Montecarlo
        A Monte Carlo sampler

    bias: BiasPotential
        A bias potential that should be altered in order to recover the free
        energy

    flat_limit: float
        The histogram of visits is considered flat, when the minimum value
        is larger than flat_limit*np.mean(hist)

    mod_factor: float
        Modification factor in units of kB*T

    fname: str
        Filename used to store the simulation state when finished
    """

    def __init__(self, mc, bias, flat_limit=0.8, mod_factor=0.1, fname="metadyn.json"):
        self.mc = mc
        self.bias = bias
        self.mc.add_bias(self.bias)
        cnst = CollectiveVariableConstraint(
            xmin=self.bias.xmin, xmax=self.bias.xmax, getter=self.bias.getter
        )
        self.mc.generator.add_constraint(cnst)
        self.mc.update_current_energy()
        self.mc.attach(self.bias.getter, interval=1)
        self.visit_hist = deepcopy(bias)
        self.visit_hist.zero()
        self.flat_limit = flat_limit
        self.mod_factor = mod_factor * kB * self.mc.T
        self.log_freq = 30
        self.fname = fname
        self.progress_info = {"mean": 0.0, "minimum": 0.0}
        self.observers = []
        self.quit = False

    def _getter_accepts_none(self):
        """Return True if the getter accepts None."""
        # TODO: Why is try-catch needed? Is it sufficient with ValueError?
        try:
            x = self.bias.getter(None)
            x = float(x)
        # pylint: disable=broad-except
        except Exception:
            return False
        return True

    def _getter_accepts_peak(self):
        """Return True if the getter supports the peak keyword."""
        # TODO: Why is try-catch needed? Is it sufficient with ValueError?
        try:
            x = self.bias.getter([], peak=True)
            x = float(x)
        # pylint: disable=broad-except
        except Exception:
            return False
        return True

    def add_observer(self, obs, interval=1):
        """Add observer.

        Parameters:

        obs: callable object
            A callable object that takes no arguments

        interval: int
            Will be called at even intervals given by this number
        """
        self.observers.append((obs, interval))

    def visit_is_flat(self):
        """Return True if the histogram of visits is flat."""
        i_min = self.visit_hist.get_index(self.visit_hist.xmin)
        i_max = self.visit_hist.get_index(self.visit_hist.xmax)
        coeff = self.visit_hist.get_coeff()[i_min:i_max]
        avg = np.mean(coeff)

        # Use min and not np.min. It looks like np.min behaves weird
        # when it is running on a worker thread
        minval = min(coeff.tolist())
        self.progress_info["mean"] = avg

        if avg > 0.0:
            self.progress_info["minval"] = minval / avg
        else:
            self.progress_info["minval"] = 0.0

        if np.max(avg) == 0:
            return False
        return minval > self.flat_limit * avg

    def update(self):
        """Update bias potential and visit histogram."""
        x = self.bias.getter(None)
        cur_value = self.bias.evaluate(x)
        self.bias.local_update(x, self.mod_factor)

        new_value = self.bias.evaluate(x)
        self.mc.current_energy += new_value - cur_value
        self.visit_hist.local_update(x, 1)

    def run(self, max_sweeps=None):
        """
        Run the calculation.

        Parameters:

        max_sweeps: int or None
            If given, the simulation terminates when this number of sweeps
            is reached
        """
        if not self._getter_accepts_none():
            raise NoneNotAcceptedError("Observer does not accept None as a system change")

        if not self._getter_accepts_peak():
            raise PeakNotAcceptedError(
                ("Observer does not accept peak as a " "keyword argument to __call__")
            )

        if not hasattr(self.bias, "get_coeff"):
            raise ValueError(
                (
                    "The bias potential needs to have a method ",
                    "called get_coeff(), which returns a histogram "
                    "representation of the bias potential",
                )
            )

        if not hasattr(self.bias, "local_update"):
            raise ValueError(
                "The bias potential needs to have a method "
                "called local_update(x, dE) which allows a local "
                "update at position x"
            )
        conv = False
        now = time.perf_counter()

        sweep_no = 0
        counter = 0
        logger.info("Starting metadynamics sampling...")
        logger.info("Writing result to %s every %s sec", self.fname, self.log_freq)
        self.mc.initialize_run()
        while not conv:
            counter += 1
            if time.perf_counter() - now > self.log_freq:
                msg = f"Sweep no. {int(counter/len(self.mc.atoms))} "
                msg += f"Average visits: {self.progress_info['mean']:.2e}. "
                msg += f"Min/avg: {self.progress_info['minval']:.2e} "
                msg += f"x: {self.bias.getter(None):.2e}"
                logger.info(msg)
                self.save()
                now = time.perf_counter()

            # pylint: disable=protected-access
            self.mc._mc_step()
            self.update()
            if self.visit_is_flat():
                conv = True

            sweep_no = int(counter / len(self.mc.atoms))

            if max_sweeps is not None:
                if sweep_no > max_sweeps:
                    logger.info("Reached max number of sweeps...")
                    conv = True

            for obs, interval in self.observers:
                if counter % (interval * len(self.mc.atoms)) == 0:
                    obs()

            if self.quit:
                break

        logger.info("Results from metadynamics sampling written to %s", self.fname)
        self.save()

    def save(self):
        """Save the free energy result to a file."""
        pot = self.bias.todict()
        xmin = self.bias.xmin
        xmax = self.bias.xmax
        x = np.linspace(xmin, xmax, 200)
        beta = 1.0 / (kB * self.mc.T)
        betaG = [self.bias.evaluate(y) * beta for y in x]
        data = {"bias_pot": pot, "betaG": {"x": x.tolist(), "y": betaG}}
        with open(self.fname, "w") as out:
            json.dump(data, out)



from clease.datastructures import SystemChanges
from .mc_constraint import MCConstraint


class PairConstraint(MCConstraint):
    """
    Pair constraint is a constraint that prevents two species
    from being part of a pair cluster

    Parameters:

    elements: list
        List of symbols (e.g. [Al, X] or [X, X])

    pair_cluster: Cluster
        Instance of a the Cluster class. An instance of a cluster class can
        for instance be obtained from a ClusterExpansionSettings object
        via settings.cluster_list.get_by_name("c2_d000_0")[0]

    trans_matrix: list of dicts
        Translation matrix for indices. This can be obtained from the
        trans_matrix attribute of the ClusterExpansionSettings object

    atoms: Atoms object
        Atoms object used for MC calculations
    """

    def __init__(self, elements, pair_cluster, trans_matrix, atoms):
        self.elements = sorted(elements)
        self.pair_cluster = pair_cluster
        self.trans_mat = trans_matrix
        self.atoms = atoms

    def _check_one(self, idx):
        t_indx = self.trans_mat[idx]
        for figure in self.pair_cluster.indices:
            symbs = sorted([self.atoms[t_indx[j]].symbol for j in figure])
            if self.elements == symbs:
                return False
        return True

    def __call__(self, system_changes: SystemChanges):
        # Introduce all changes
        for change in system_changes:
            self.atoms[change.index].symbol = change.new_symb

        # Loop through all changes
        move_ok = True
        for change in system_changes:
            if not self._check_one(change.index):
                move_ok = False
                break

        # Undo all changes
        for change in system_changes:
            self.atoms[change.index].symbol = change.old_symb
        return move_ok



# pylint: disable=undefined-variable
from .mc_observer import MCObserver
from .sgc_observer import SGCObserver
from .corr_func_observer import CorrelationFunctionObserver
from .snapshot_observer import Snapshot
from .energy_evolution_obs import EnergyEvolution
from .site_order_parameter import SiteOrderParameter
from .lowest_energy_observer import LowestEnergyStructure
from .diffraction_observer import DiffractionObserver
from .energy_plot_updater import EnergyPlotUpdater
from .concentration_observer import ConcentrationObserver
from .multi_state_sgc_observer import MultiStateSGCConcObserver, SGCState
from .entropy_prod_rate import *
from .acceptance_rate import *
from .move_obs import *

ADDITIONAL = (
    "MCObserver",
    "SGCObserver",
    "CorrelationFunctionObserver",
    "Snapshot",
    "EnergyEvolution",
    "LowestEnergyStructure",
    "DiffractionObserver",
    "EnergyPlotUpdater",
    "ConcentrationObserver",
    "SiteOrderParameter",
    "MultiStateSGCConcObserver",
    "SGCState",
)

__all__ = ADDITIONAL + (entropy_prod_rate.__all__ + acceptance_rate.__all__ + move_obs.__all__)



import attr
from clease.jsonio import AttrSavable, jsonable
from .system_changes import SystemChanges

__all__ = ("MCStep",)


def _compare_moves(move1: SystemChanges, move2: SystemChanges) -> bool:
    """Helper function to compare equality between system changes."""
    if len(move1) != len(move2):
        return False
    for change1, change2 in zip(move1, move2):
        if change1 != change2:
            return False
    return True


@jsonable("mc_step")
@attr.define(eq=True, slots=True)
class MCStep(AttrSavable):
    """Container with information about a single MC step.
    No validation checks are made in this class for performance reasons.
    """

    step: int = attr.field()
    energy: float = attr.field()
    move_accepted: bool = attr.field()
    # last_move may be a tuple or list for example, so ensure we can compare them
    last_move: SystemChanges = attr.field(
        eq=attr.cmp_using(eq=_compare_moves, require_same_type=False)
    )

    @property
    def move_rejected(self) -> bool:
        return not self.move_accepted



import numpy as np
from scipy.spatial import ConvexHull as SciConvexHull
import matplotlib.pyplot as plt
from ase.db import connect
from clease.tools import invert_matrix

__all__ = ("ConvexHull",)


class ConvexHull:
    """
    Evaluates data from database and construct the convex hull

    Parameters:

    db_name: str
        Name of the database

    select_cond: list
        Select conditions for retrieving data from the database
        If None, the select condition will be [[('converged', '=', True)]

    atoms_per_fu: int
        Number of atoms per formula unit

    conc_scale: float
        Concentration scale. In the plots the concentration
        will be divided by this number

    conc_ranges: dict
        Dictionary with lower and upper bound for the concentrations to
        be included.
        Example:
        If one want to limit the evaluation to configurations having
        gold concentrations in the range 0 to 0.5 this argument would
        be {"Au": (0, 0.5)}.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        db_name,
        select_cond=None,
        atoms_per_fu=1,
        conc_scale=1.0,
        conc_ranges: dict = None,
    ):
        if conc_ranges is None:
            conc_ranges = {}
        self.db_name = db_name
        self.atoms_per_fu = atoms_per_fu
        self.conc_scale = conc_scale
        self.conc_ranges = conc_ranges
        self.db = connect(self.db_name)

        if select_cond is None:
            self.select_cond = [("converged", "=", True)]
        else:
            self.select_cond = select_cond

        self._unique_elem = sorted(list(self.unique_elements()))
        self.end_points = self._get_end_points()
        self.weights = self._weighting_coefficients(self.end_points)
        self.num_varying_concs = 1

        self.energies, self.concs, self.db_ids = self._get_conc_energies()

    def unique_elements(self):
        """Return the number of unique elements."""
        elems = set()
        for row in self.db.select(self.select_cond):
            if not self._include_row(row):
                continue
            count = row.count_atoms()
            elems = elems.union(set(count.keys()))
        return elems

    def _include_row(self, row):
        """Return True if data from the row should be included."""

        atoms_count = row.count_atoms()
        for k in atoms_count.keys():
            atoms_count[k] /= row.natoms
        for k, v in self.conc_ranges.items():
            conc = atoms_count.get(k, 0.0)
            if conc < v[0] or conc > v[1]:
                return False
        return True

    def _get_end_points(self):
        """Return the end points based on information in the database

        The algorithm seeks one configuration that maximize
        the composition of each element in the database.
        """
        # Initialize the end-point dictionaries
        end_points = {k: {} for k in self._unique_elem}
        for v in end_points.values():
            for k2 in self._unique_elem:
                v[f"{k2}_conc"] = 0.0
            v["energy"] = np.infty

        # Iterate through all rows, update
        # by finding the lowest energy structure,
        # which has the maximum amount of a given element
        for row in self.db.select(self.select_cond):
            if not self._include_row(row):
                continue
            conc = row.count_atoms()
            # Build the concentration dictionary for this row
            for k in self._unique_elem:
                if k not in conc:
                    conc[k] = 0.0
                else:
                    conc[k] /= row.natoms

            # Find out if the structure is "better" for any of the
            # end point elements
            for element, v in end_points.items():
                # Concentration of the current element to be maximized
                own_conc = v[f"{element}_conc"]
                # Check if the current structure
                # is an endpoint
                if conc[element] >= own_conc:
                    energy = _get_row_energy(row, self.db)
                    if conc[element] == own_conc and energy >= v["energy"]:
                        # We already have a record with the same concentration,
                        # but a better (lower) energy
                        continue
                    # Either a higher concentration or a lower energy
                    # update our current best
                    v["energy"] = energy
                    for elem, new_conc in conc.items():
                        v[f"{elem}_conc"] = new_conc

        return end_points

    def _weighting_coefficients(self, end_points):
        """Return a dictionary with coefficient on reference
           energy should be weighted.

        The weights are constructed as follows:
        1. The formation energies of each end point should
            be zero
        2. To obtain the formation energy of an arbitrary
            structure, one should subtract the inner product
            between the concentration and the weights

        Parameters:

        end_points: dict
            Dictionary with end point information (internally calculated)
        """
        matrix = np.zeros((len(end_points), len(self._unique_elem)))
        rhs = np.zeros(len(end_points))
        row = 0
        for _, v in end_points.items():
            for j, symb in enumerate(self._unique_elem):
                matrix[row, j] = v[f"{symb}_conc"]
            rhs[row] = v["energy"]
            row += 1

        inv_mat = invert_matrix(matrix)

        coeff = inv_mat.dot(rhs)
        weights = {s: coeff[i] for i, s in enumerate(self._unique_elem)}
        return weights

    def _get_conc_energies(self):
        """Read concentrations, energy and ID from the database."""
        energies = []
        ids = []
        conc = {k: [] for k in self._unique_elem}

        for row in self.db.select(self.select_cond):
            if not self._include_row(row):
                continue
            count = row.count_atoms()

            for k in conc.keys():
                if k not in count.keys():
                    conc[k].append(0.0)
                else:
                    conc[k].append(count[k] / row.natoms)

            final_struct_id = row.get("final_struct_id", -1)
            if final_struct_id >= 0:
                # New format where energy is stored in a separate DB entry
                form_energy = self.db.get(id=final_struct_id).energy / row.natoms
            else:
                # Old format where the energy is stored in the init structure
                form_energy = row.energy / row.natoms

            # Subtract the appropriate weights
            form_energy -= sum(conc[k][-1] * self.weights[k] for k in conc.keys())
            energies.append(form_energy)
            ids.append(row.id)

        return np.array(energies), conc, ids

    def get_convex_hull(self, conc_var=None) -> SciConvexHull:
        """Return the convex hull.

        Parameters:

        conc_var: str
            Concentration variable used when calculating the
            convex hull.
        """

        if conc_var is None:
            num_comp = len(self._unique_elem) - 1
            x = np.zeros((len(self.energies), num_comp))
            elems = list(self._unique_elem)
            for i in range(num_comp):
                x[:, i] = self.concs[elems[i]]
        elif conc_var in self._unique_elem:
            x = np.array(self.concs[conc_var])
        else:
            msg = f"conc_var has to be {self._unique_elem} or None"
            raise ValueError(msg)

        points = np.vstack((x.T, self.energies.T)).T
        conv_hull = SciConvexHull(points)
        return conv_hull

    def _is_lower_conv_hull(self, simplex):
        """Return True if the simplex contain points that are
            non-positive.

        Parameters:

        simplex: list
            List with indices of the points that lies
            on the convex hull
        """
        tol = 1e-4
        return all(self.energies[i] <= tol for i in simplex)

    def plot(self, fig=None, concs=None, energies=None, marker="o", mfc="none"):
        """Plot formation energies."""
        # pylint: disable=too-many-branches
        # We only add the Convex Hull for the DFT
        # data
        add_cnv_hull = concs is None
        num_plots = len(self._unique_elem) - 1

        # Figure out how many plots we need
        varying_concs = []
        for k in self._unique_elem:
            minconc = np.min(self.concs[k])
            maxconc = np.max(self.concs[k])
            if abs(maxconc - minconc) > 0.01:
                varying_concs.append(k)

        num_plots = len(varying_concs) - 1

        if fig is None:
            fig = plt.figure()
            for i in range(num_plots):
                fig.add_subplot(1, num_plots, i + 1)

        if concs is None:
            concs = self.concs
        if energies is None:
            energies = self.energies

        elems = sorted(varying_concs)[:-1]
        for i, ax in enumerate(fig.get_axes()):
            # ax = fig.add_subplot(1, num_plots, i+1)

            x = np.array(concs[elems[i]])
            x /= self.conc_scale

            ax.plot(x, np.asarray(energies) * self.atoms_per_fu, marker, mfc=mfc)

            if self.atoms_per_fu > 1:
                unit = "eV/f.u."
            else:
                unit = "eV/atom"
            if i == 0:
                ax.set_ylabel(f"Formation energy ({unit})")
            else:
                ax.set_yticklabels([])

            if add_cnv_hull:
                c_hull = self.get_convex_hull(conc_var=elems[i])
                # pylint cannot inspect scipy C things, false positive
                # pylint: disable=no-member
                for simpl in c_hull.simplices:
                    if self._is_lower_conv_hull(simpl):
                        x_cnv = [x[simpl[0]], x[simpl[1]]]
                        y_cnv = [
                            self.energies[simpl[0]] * self.atoms_per_fu,
                            self.energies[simpl[1]] * self.atoms_per_fu,
                        ]
                        ax.plot(x_cnv, y_cnv, color="black")
            ax.set_xlabel(f"{elems[i]} conc")
        return fig

    def show_structures_on_convex_hull(self):
        """Show all entries on the convex hull."""
        # pylint: disable=import-outside-toplevel
        from ase.gui.gui import GUI
        from ase.gui.images import Images

        c_hull = self.get_convex_hull()
        indices = set()
        # pylint cannot inspect scipy C things, false positive
        # pylint: disable=no-member
        for simplex in c_hull.simplices:
            if self._is_lower_conv_hull(simplex):
                indices = indices.union(simplex)

        cnv_hull_atoms = []
        for i in indices:
            db_id = self.db_ids[i]
            atoms = self.db.get(id=db_id).toatoms()
            cnv_hull_atoms.append(atoms)

        images = Images()
        images.initialize(cnv_hull_atoms)
        gui = GUI(images, expr="")
        gui.run()

    def cosine_similarity_convex_hull(self, conc, tot_en, cnv_hull=None):
        """Calculate the Cosine similarity with structures on the convex hull.

        Parameters:

        conc: dict
            Dictionary with concentrations. If the system consists of
            0.3 Au, 0.5 Cu and 0.2 Zn, this would be
            {'Au': 0.3, 'Cu': 0.5, 'Zn': 0.2}
        tot_en: float
            Total energy per atom
        cnv_hull: scipy.spatial.ConvexHull
            Convex hull object holding the data.
        """

        if cnv_hull is None:
            cnv_hull = self.get_convex_hull()

        indices = set()
        for simplex in cnv_hull.simplices:
            if self._is_lower_conv_hull(simplex):
                indices = indices.union(simplex)

        data = cnv_hull.points[list(indices), :]

        form_energy = tot_en - sum(self.weights[k] * conc[k] for k in conc.keys())

        min_en = np.min(data[:, -1])
        max_en = np.max(data[:, -1])
        diff = max_en - min_en
        normalization = 0.5 * diff

        # Normalize the energy data
        data[:, -1] /= normalization

        mean = np.mean(data, axis=0)
        data -= mean

        data_vec = np.zeros(data.shape[1])
        data_vec[:-1] = [conc.get(k, 0.0) for k in self._unique_elem[:-1]]
        data_vec[-1] = form_energy / normalization
        data_vec -= mean

        inner_prod = data.dot(data_vec)
        inner_prod /= np.sqrt(data_vec.dot(data_vec))
        inner_prod /= np.sqrt(np.diag(data.dot(data.T)))
        return np.max(inner_prod)

    def get_formation_energy(self, conc, tot_energy):
        """Return the formation energy

        Parameters:

        conc: dict
            Dictionary with the concenatration
        tot_energy: float
            Total energy per atom
        """
        return tot_energy - sum(self.weights[k] * conc[k] for k in conc.keys())


def _get_row_energy(row, con) -> float:
    f_id = row.get("final_struct_id", -1)
    if f_id >= 0:
        # New format where energy is in a separate entry
        return con.get(id=f_id).energy / row.natoms
    # Old format where the energy is stored in same entry
    return row.energy / row.natoms



from typing import Dict
import ase
from clease.datastructures import MCStep, SystemChanges
from clease.montecarlo.averager import Averager
from .mc_observer import MCObserver


class ConcentrationObserver(MCObserver):
    """
    Observer that can be attached to a MC run, to track the concenctration of a
    particular element. This observer has to be executed on every MC step.

    Parameters:

    atoms: Atoms object
        Atoms object used for MC

    element: str
        The element that should be tracked
    """

    name = "ConcentrationObserver"

    def __init__(self, atoms: ase.Atoms, element: str):
        super().__init__()
        self.element = element
        self.n = len(atoms)
        self.init_conc = self.calculate_from_scratch(atoms)
        self.current_conc = self.init_conc
        self._make_averagers()

    def new_concentration(self, system_changes: SystemChanges) -> float:
        """Calculate the new consentration after the changes."""
        new_conc = self.current_conc
        for change in system_changes:
            if change.new_symb == self.element:
                new_conc += 1.0 / self.n
            if change.old_symb == self.element:
                new_conc -= 1.0 / self.n
        return new_conc

    def __call__(self, system_changes: SystemChanges, peak: bool = False) -> float:
        """Implement the __call__ method to work with the BiasPotentials"""
        if not system_changes:
            # No changes, None or some other falsey thing.
            return self.current_conc

        new_conc = self.new_concentration(system_changes)
        if peak:
            return new_conc

        self.current_conc = new_conc
        self.avg_conc += new_conc
        self.avg_conc_sq += new_conc**2
        return self.current_conc

    def observe_step(self, mc_step: MCStep, peak: bool = False) -> float:
        if mc_step.move_rejected:
            return self.current_conc

        return self(mc_step.last_move, peak=peak)

    def reset(self) -> None:
        """Reset the observer"""
        # Remake new average objects, since we also change the reference concentration
        self._make_averagers()

    def _make_averagers(self) -> None:
        """Construct the internal averager objects. Starts with the current concentration
        as the first sample."""
        self.avg_conc = Averager(ref_value=self.current_conc)
        self.avg_conc_sq = Averager(ref_value=self.current_conc**2)
        # The current concentration is the first sample
        self.avg_conc += self.current_conc
        self.avg_conc_sq += self.current_conc**2

    def get_averages(self) -> Dict[str, float]:
        mean_conc = self.avg_conc.mean
        var_conc = self.avg_conc_sq.mean - mean_conc**2
        return {f"conc_{self.element}": mean_conc, f"conc_var_{self.element}": var_conc}

    def calculate_from_scratch(self, atoms: ase.Atoms) -> float:
        """Calculate the concentration of the element in the atoms object."""
        num_atoms = sum(atoms.symbols == self.element)
        return num_atoms / len(atoms)

    def interval_ok(self, interval: int) -> bool:
        """Every step must be observed, as otherwise we'd miss updates,
        and the concentration becomes incorerct."""
        return interval == 1



"""Module for defining deprecated regression class imports"""
# pylint: skip-file
from deprecated import deprecated
from clease import regression

__all__ = (
    "LinearRegression",
    "Tikhonov",
    "Lasso",
    "BayesianCompressiveSensing",
    "ConstrainedRidge",
    "GAFit",
    "GeneralizedRidgeRegression",
    "PhysicalRidge",
    "SequentialClusterRidge",
)

MSG = "Import {} from clease.regression instead"
DEP_VERSION = "0.10.0"  # Deprecation version


@deprecated(version=DEP_VERSION, reason=MSG.format("LinearRegression"))
class LinearRegression(regression.LinearRegression):
    pass


@deprecated(version=DEP_VERSION, reason=MSG.format("Tikhonov"))
class Tikhonov(regression.Tikhonov):
    pass


@deprecated(version=DEP_VERSION, reason=MSG.format("Lasso"))
class Lasso(regression.Lasso):
    pass


@deprecated(version=DEP_VERSION, reason=MSG.format("BayesianCompressiveSensing"))
class BayesianCompressiveSensing(regression.BayesianCompressiveSensing):
    pass


@deprecated(version=DEP_VERSION, reason=MSG.format("ConstrainedRidge"))
class ConstrainedRidge(regression.ConstrainedRidge):
    pass


@deprecated(version=DEP_VERSION, reason=MSG.format("GAFit"))
class GAFit(regression.GAFit):
    pass


@deprecated(version=DEP_VERSION, reason=MSG.format("GeneralizedRidgeRegression"))
class GeneralizedRidgeRegression(regression.GeneralizedRidgeRegression):
    pass


@deprecated(version=DEP_VERSION, reason=MSG.format("PhysicalRidge"))
class PhysicalRidge(regression.PhysicalRidge):
    pass


@deprecated(version=DEP_VERSION, reason=MSG.format("SequentialClusterRidge"))
class SequentialClusterRidge(regression.SequentialClusterRidge):
    pass



# pylint: disable=undefined-variable
from .cluster_fingerprint import *
from .cluster import *
from .cluster_generator import *
from .cluster_info_mapper import *
from .cluster_list import *
from .cluster_manager import *

from . import utils

ADDITIONAL = ("utils",)

__all__ = (
    cluster.__all__
    + cluster_fingerprint.__all__
    + cluster_generator.__all__
    + cluster_info_mapper.__all__
    + cluster_list.__all__
    + cluster_manager.__all__
    + ADDITIONAL
)



import numpy as np
from clease.datastructures import SystemChanges
from .mc_observer import MCObserver


class DiffractionUpdater:
    """
    Utility class for all objects that require tracing of Fourier reflection.
    See docstring of DiffractionObserver for explanation of the arguments.
    This observer has to be executed on every MC step.
    """

    def __init__(self, atoms=None, k_vector=(), active_symbols=(), all_symbols=()):
        self.orig_symbols = [atom.symbol for atom in atoms]
        self.k_vector = k_vector
        self.N = len(atoms)
        self.k_dot_r = atoms.get_positions().dot(self.k_vector)
        self.indicator = {k: 0 for k in all_symbols}
        for symb in active_symbols:
            self.indicator[symb] = 1.0

        self.value = self.calculate_from_scratch(self.orig_symbols)
        self.prev_value = self.value

    def update(self, system_changes: SystemChanges):
        """Update the reflection value."""
        self.prev_value = self.value
        for change in system_changes:
            f_val = np.exp(1j * self.k_dot_r[change.index]) / self.N
            self.value += self.indicator[change.new_symb] * f_val
            self.value -= self.indicator[change.old_symb] * f_val

    def undo(self):
        """Undo the last update."""
        self.value = self.prev_value

    def reset(self):
        """Reset all values."""
        self.value = self.calculate_from_scratch(self.orig_symbols)
        self.prev_value = self.value

    def calculate_from_scratch(self, symbols):
        """Calculate the intensity from sctrach."""
        value = 0.0 + 1j * 0.0
        for i, symb in enumerate(symbols):
            value += self.indicator[symb] * np.exp(1j * self.k_dot_r[i])
        return value / len(symbols)


class DiffractionObserver(MCObserver):
    """
    Trace the reflection intensity.

    Parameters:

    atoms: Atoms
        Atoms object used in Monte Carlo

    k_vector: list
        Fourier reflection to be traced

    active_symbols: list
        List of symbols that reflects

    all_symbols: list
        List of all symbols in the simulation

    name: str
        Name of the DiffractionObserver
        (users are given the freedom to set names because they can attach
        multiple DiffractionObserver instances)


    Example:

    Consider a system where Al, Mg and Si occupy FCC lattice sites. We want to
    trace the occurence of Mg layers that are separated by a distance 3*a
    where *a* is the lattice parameter. We further assume that the *y*-axis is
    normal to the planes we want to trace. In that case, we specify the
    variables as

    >>> from ase.build import bulk
    >>> import numpy as np
    >>> a = 4.05
    >>> atoms = bulk('Al', crystalstructure='fcc', a=a)
    >>> k_vector = [0, 2.0*np.pi/(3*a), 0]
    >>> active_elements = ['Mg']
    >>> all_symbols = ['Al', 'Mg', 'Si']

    If we do not wish to distinguish Mi and Si (we do not distiguish Mg layer,
    Si layer or a mixture of the two) the `active_elements` is changed to

    >>> active_elements = ['Mg', 'Si']
    """

    def __init__(
        self,
        atoms=None,
        k_vector=(),
        active_symbols=(),
        all_symbols=(),
        name="reflection1",
    ):
        super().__init__()
        self.updater = DiffractionUpdater(
            atoms=atoms,
            k_vector=k_vector,
            active_symbols=active_symbols,
            all_symbols=all_symbols,
        )
        self.avg = self.updater.value
        self.num_updates = 1
        self.name = name

    def __call__(self, system_changes: SystemChanges):
        self.updater.update(system_changes)
        self.avg += self.updater.value

    def get_averages(self):
        return {self.name: np.abs(self.avg / self.num_updates)}

    def reset(self):
        self.updater.reset()
        self.avg = self.updater.value
        self.num_updates = 1

    def interval_ok(self, interval):
        return interval == 1



from typing import Union
import logging
from ase import Atoms
from clease.datastructures import SystemChanges
from clease.calculator import Clease
from clease.settings.settings import ClusterExpansionSettings

__all__ = ("MCEvaluator", "CEMCEvaluator", "construct_evaluator")

logger = logging.getLogger(__name__)


class MCEvaluator:
    """A Montecarlo evaluator class, used to perform the energy evaluations
    within a Montecarlo run.

    Args:
        atoms (ase.Atoms): ASE Atoms object to be used for the evaluation.
            This atoms object may be mutated.
    """

    def __init__(self, atoms: Atoms):
        self._atoms = atoms

    @property
    def atoms(self):
        return self._atoms

    def get_energy(self, applied_changes: SystemChanges = None) -> float:
        """Evaluate the energy of a system.
        If a change is sufficiently local/small, it there, in some situations,
        may be other ways of evaluating the energy than a full calculation.
        Must return the energy of the new configuration.

        The applied changes only reflect what has already been applied to the system.

        Args:
            applied_changes (SystemChanges, optional): A list of changes which has been applied to
                the atoms object. This change has already been applied, and is only for bookkeeping
                purposes, if evaluation schemas want to make decisions based on what has changed.
                Defaults to None.

        Returns:
            float: Energy of the atoms object.
        """
        # pylint: disable=unused-argument
        # system_changes are passed in (optionally) in order to allow for localized evaluations
        # See discussion:
        # https://gitlab.com/computationalmaterials/clease/-/issues/268
        return self.atoms.get_potential_energy()

    def reset(self) -> None:
        """Perform a reset on the evaluator and/or on the atoms"""

    def apply_system_changes(self, system_changes: SystemChanges, keep=False) -> None:
        """Mutate the atoms object to reflect the system change.

        Args:
            system_changes (SystemChanges): Sequence of changes to be applied.
            keep (bool, optional): Whether to call
                :meth:`~clease.montecarlo.mc_evaluator.MCEvaluator.keep_system_changes`
                after applying changes. Defaults to False.
        """
        for change in system_changes:
            change.apply_change(self.atoms)
        if keep:
            self.keep_system_changes(system_changes=system_changes)

    def undo_system_changes(self, system_changes: SystemChanges) -> None:
        """Mutate the atoms object to undo the system change.

        Args:
            atoms (Atoms): Atoms object to be mutated.
            system_changes (SystemChanges): Sequence of changes to be applied.
        """
        for change in system_changes:
            change.undo_change(self.atoms)

    def keep_system_changes(self, system_changes: SystemChanges = None) -> None:
        """A set of system changes are to be kept. Perform necessary actions to prepare
        for a new evaluation."""

    def get_energy_given_change(self, system_changes: SystemChanges) -> float:
        """Calculate the energy of a set of changes, and undo any changes.

        Args:
            atoms (Atoms): Atoms object to be mutated.
            system_changes (SystemChanges): Sequence of changes to be applied.

        Returns:
            float: The resulting energy from a
            call to :meth:`~clease.montecarlo.mc_evaluator.MCEvaluator.get_energy`.
        """
        self.apply_system_changes(system_changes)
        energy = self.get_energy(applied_changes=system_changes)
        self.undo_system_changes(system_changes)
        return energy

    def synchronize(self) -> None:
        """Ensure the calculator and atoms objects are synchronized."""


class CEMCEvaluator(MCEvaluator):
    """MC Evaluator to be used with the CLEASE CE Calculator.
    Assumes the attached calculator is a Clease Calculator"""

    def __init__(self, atoms: Atoms):
        if not isinstance(atoms.calc, Clease):
            raise ValueError(
                "Clease calculator must be attached to the atoms object " "when using CEMCEvaluator"
            )
        super().__init__(atoms)

    @property
    def settings(self) -> ClusterExpansionSettings:
        """Get the related settings object"""
        return self.calc.settings

    @property
    def calc(self) -> Clease:
        """Get the internal calculator object"""
        return self.atoms.calc

    def get_energy(self, applied_changes: SystemChanges = None) -> float:
        return self.calc.get_energy()

    def reset(self) -> None:
        """Perform a reset on the evaluator and/or on the atoms"""
        self.calc.clear_history()
        super().reset()

    def apply_system_changes(self, system_changes: SystemChanges, keep=False) -> None:
        """Mutate the atoms object to reflect the system change.

        Args:
            atoms (Atoms): Atoms object to be mutated.
            system_changes (SystemChanges): Sequence of changes to be applied.
            keep (bool, optional): Whether to call
                :meth:`~clease.montecarlo.mc_evaluator.MCEvaluator.keep_system_changes`
                after applying changes. Defaults to False.
        """
        self.calc.apply_system_changes(system_changes)
        if keep:
            self.keep_system_changes(system_changes=system_changes)

    def undo_system_changes(self, system_changes: SystemChanges) -> None:
        """Mutate the atoms object to undo the system change.

        Args:
            atoms (Atoms): Atoms object to be mutated.
            system_changes (SystemChanges): Sequence of changes to be applied.
        """
        self.calc.undo_system_changes()

    def keep_system_changes(self, system_changes: SystemChanges = None) -> None:
        """A set of system changes are to be kept. Perform necessary actions to prepare
        for a new evaluation."""
        self.calc.keep_system_changes()

    def get_energy_given_change(self, system_changes: SystemChanges) -> float:
        """Calculate the energy of a set of changes, and undo any changes.

        Args:
            atoms (Atoms): Atoms object to be mutated.
            system_changes (SystemChanges): Sequence of changes to be applied.

        Returns:
            float: The resulting energy from a
            call to :meth:`~clease.montecarlo.mc_evaluator.MCEvaluator.get_energy`.
        """
        return self.calc.get_energy_given_change(system_changes)

    def synchronize(self) -> None:
        """Ensure the calculator and atoms objects are synchronized."""
        # Recalculate the CF
        self.calc.update_cf()


def _make_mc_evaluator_from_atoms(atoms: Atoms) -> MCEvaluator:
    """Construct a new MC Evaluator based on the calculator object attached to the Atoms object.

    Raises a ``RuntimeError`` is the Atoms object has no calculator."""
    calc = atoms.calc
    if calc is None:
        raise RuntimeError("Atoms object must have a calculator object.")
    if isinstance(calc, Clease):
        # Return the MC Evaluator specialized for the CLEASE calculator
        logger.debug("Constructed a new CE MC Evaluator.")
        return CEMCEvaluator(atoms)
    # Return the generic MC Evaluator
    logger.debug("Constructed a generic MC Evaluator.")
    return MCEvaluator(atoms)


def construct_evaluator(system: Union[Atoms, MCEvaluator]) -> MCEvaluator:
    """Helper function for constructing a new evaluator object, either by passing
    in an ASE atoms object or an explicit evaluator object.

    Args:
        system (Union[Atoms, MCEvaluator]): If the system is an Atoms object,
            then a new :class:`~clease.montecarlo.mc_evaluator.MCEvaluator`
            object is created on the basis of the attached calculator.
            Otherwise, if the system is already an Evaluator object,
            the same object is returned.

    Raises:
        TypeError: If the system is not either an ASE atoms object
            or an ``MCEvaluator``.

    Returns:
        MCEvaluator: The evaluator object to be used.
    """

    if isinstance(system, Atoms):
        # Construct the default evaluator object
        logger.debug("Creating a new default evaluator.")
        return _make_mc_evaluator_from_atoms(system)
    if isinstance(system, MCEvaluator):
        # User supplied a custom evaluator
        logger.debug("Using a user-defined evaluator: %s", system)
        return system
    raise TypeError(f"Received an unknown system: {system}")



from pathlib import Path
from packaging.version import parse

with Path(__file__).with_name("_version.txt").open("r") as f:
    __version__ = parse(f.readline().strip())

# Representation of version_info as (x, y, z)
version_info = __version__.release

__all__ = ("__version__", "version_info")
